import collections

import pytest

from textacy import utils
from textacy.datasets import base


DATASET = base.Dataset("foo", {"test": True})


def test_repr():
    assert utils.to_unicode(str(DATASET)).startswith("Dataset")


def test_info():
    assert isinstance(DATASET.info, dict)
    assert "name" in DATASET.info
    assert "test" in DATASET.info


def test_iter():
    assert isinstance(DATASET, collections.abc.Iterable)


def test_methods():
    for name in ("texts", "records", "download"):
        assert hasattr(DATASET, name)
        with pytest.raises(NotImplementedError):
            getattr(DATASET, name)()



from __future__ import annotations

import collections
from operator import itemgetter
from typing import Callable, Collection, Dict, List, Optional, Set, Tuple

from spacy.tokens import Doc, Token

from ... import representations, utils
from .. import utils as ext_utils


def textrank(
    doc: Doc,
    *,
    normalize: Optional[str | Callable[[Token], str]] = "lemma",
    include_pos: Optional[str | Collection[str]] = ("NOUN", "PROPN", "ADJ"),
    window_size: int = 2,
    edge_weighting: str = "binary",
    position_bias: bool = False,
    topn: int | float = 10,
) -> List[Tuple[str, float]]:
    """
    Extract key terms from a document using the TextRank algorithm, or
    a variation thereof. For example:

    - TextRank: ``window_size=2, edge_weighting="binary", position_bias=False``
    - SingleRank: ``window_size=10, edge_weighting="count", position_bias=False``
    - PositionRank: ``window_size=10, edge_weighting="count", position_bias=True``

    Args:
        doc: spaCy ``Doc`` from which to extract keyterms.
        normalize: If "lemma", lemmatize terms; if "lower", lowercase terms; if None,
            use the form of terms as they appeared in ``doc``; if a callable,
            must accept a ``Token`` and return a str,
            e.g. :func:`textacy.spacier.utils.get_normalized_text()`.
        include_pos: One or more POS tags with which to filter for good candidate keyterms.
            If None, include tokens of all POS tags
            (which also allows keyterm extraction from docs without POS-tagging.)
        window_size: Size of sliding window in which term co-occurrences are determined.
        edge_weighting ({"count", "binary"}): : If "count", the nodes for
            all co-occurring terms are connected by edges with weight equal to
            the number of times they co-occurred within a sliding window;
            if "binary", all such edges have weight = 1.
        position_bias: If True, bias the PageRank algorithm for weighting
            nodes in the word graph, such that words appearing earlier and more
            frequently in ``doc`` tend to get larger weights.
        topn: Number of top-ranked terms to return as key terms.
            If an integer, represents the absolute number; if a float, value
            must be in the interval (0.0, 1.0], which is converted to an int by
            ``int(round(len(set(candidates)) * topn))``.

    Returns:
        Sorted list of top ``topn`` key terms and their corresponding TextRank ranking scores.

    References:
        - Mihalcea, R., & Tarau, P. (2004, July). TextRank: Bringing order into texts.
          Association for Computational Linguistics.
        - Wan, Xiaojun and Jianguo Xiao. 2008. Single document keyphrase extraction
          using neighborhood knowledge. In Proceedings of the 23rd AAAI Conference
          on Artificial Intelligence, pages 855–860.
        - Florescu, C. and Cornelia, C. (2017). PositionRank: An Unsupervised Approach
          to Keyphrase Extraction from Scholarly Documents. In proceedings of ACL*,
          pages 1105-1115.
    """
    # validate / transform args
    include_pos = utils.to_collection(include_pos, str, set)
    if isinstance(topn, float):
        if not 0.0 < topn <= 1.0:
            raise ValueError(
                "topn={} is invalid; "
                "must be an int, or a float between 0.0 and 1.0".format(topn)
            )

    # bail out on empty docs
    if not doc:
        return []

    word_pos: Optional[Dict[str, float]]
    if position_bias is True:
        word_pos = collections.defaultdict(float)
        for word, norm_word in zip(doc, ext_utils.terms_to_strings(doc, normalize)):
            word_pos[norm_word] += 1 / (word.i + 1)
        sum_word_pos = sum(word_pos.values())
        word_pos = {word: pos / sum_word_pos for word, pos in word_pos.items()}
    else:
        word_pos = None
    # build a graph from all words in doc, then score them
    graph = representations.network.build_cooccurrence_network(
        list(ext_utils.terms_to_strings(doc, normalize)),
        window_size=window_size,
        edge_weighting=edge_weighting,
    )
    word_scores = representations.network.rank_nodes_by_pagerank(
        graph, weight="weight", personalization=word_pos
    )
    # generate a list of candidate terms
    candidates = _get_candidates(doc, normalize, include_pos)
    if isinstance(topn, float):
        topn = int(round(len(set(candidates)) * topn))
    # rank candidates by aggregating constituent word scores
    candidate_scores = {
        " ".join(candidate): sum(word_scores.get(word, 0.0) for word in candidate)
        for candidate in candidates
    }
    sorted_candidate_scores = sorted(
        candidate_scores.items(), key=itemgetter(1, 0), reverse=True
    )
    return ext_utils.get_filtered_topn_terms(
        sorted_candidate_scores, topn, match_threshold=0.8
    )


def _get_candidates(
    doc: Doc, normalize: Optional[str | Callable], include_pos: Optional[Set[str]],
) -> Set[Tuple[str, ...]]:
    """
    Get a set of candidate terms to be scored by joining the longest
    subsequences of valid words -- non-stopword and non-punct, filtered to
    nouns, proper nouns, and adjectives if ``doc`` is POS-tagged -- then
    normalized into strings.
    """

    def _is_valid_tok(tok):
        return not (tok.is_stop or tok.is_punct or tok.is_space) and (
            include_pos is None or tok.pos_ in include_pos
        )

    candidates = ext_utils.get_longest_subsequence_candidates(doc, _is_valid_tok)
    return {
        tuple(ext_utils.terms_to_strings(candidate, normalize)) for candidate in candidates
    }



import pytest

import textacy
import textacy.resources


RESOURCE = textacy.resources.DepecheMood(lang="en", word_rep="lemmapos")

pytestmark = pytest.mark.skipif(
    RESOURCE.filepath is None,
    reason="DepecheMood resource must be downloaded before running tests",
)


@pytest.fixture(scope="module")
def spacy_doc():
    text = "A friend is someone who knows all about you and still loves you."
    return textacy.make_spacy_doc(text, lang="en_core_web_sm")


@pytest.mark.skip("No need to download a new resource every time")
def test_download(tmpdir):
    resource = textacy.resources.DepecheMood(data_dir=str(tmpdir))
    resource.download()
    assert resource._filepath.is_file()


def test_oserror(tmpdir):
    resource = textacy.resources.DepecheMood(data_dir=str(tmpdir))
    with pytest.raises(OSError):
        _ = resource.weights


class TestInitParams:

    def test_lang(self):
        for lang in ["en", "it"]:
            _ =  textacy.resources.DepecheMood(lang=lang, word_rep="token")
        for lang in ["es", None]:
            with pytest.raises(ValueError):
                _ =  textacy.resources.DepecheMood(lang=lang)

    def test_word_rep(self):
        for word_rep in ["token", "lemma"]:
            _ =  textacy.resources.DepecheMood(lang="it", word_rep=word_rep)
        for word_rep in ["lower", None]:
            with pytest.raises(ValueError):
                _ =  textacy.resources.DepecheMood(word_rep=word_rep)


class TestGetEmotionalVariance:

    def test_str(self):
        result = RESOURCE.get_emotional_valence("love#n")
        assert isinstance(result, dict)
        assert all(isinstance(key, str) for key in result.keys())
        assert all(isinstance(val, float) for val in result.values())
        assert all(0 <= val <= 1.0 for val in result.values())

    def test_sequence_str(self):
        result = RESOURCE.get_emotional_valence(["love#n", "hate#n"])
        assert isinstance(result, dict)
        assert all(isinstance(key, str) for key in result.keys())
        assert all(isinstance(val, float) for val in result.values())
        assert all(0 <= val <= 1.0 for val in result.values())

    def test_token(self, spacy_doc):
        for tok in spacy_doc:
            result = RESOURCE.get_emotional_valence(tok)
            assert isinstance(result, dict)
            assert all(isinstance(key, str) for key in result.keys())
            assert all(isinstance(val, float) for val in result.values())
            assert all(0 <= val <= 1.0 for val in result.values())

    def test_span(self, spacy_doc):
        result = RESOURCE.get_emotional_valence(spacy_doc[:8])
        assert isinstance(result, dict)
        assert all(isinstance(key, str) for key in result.keys())
        assert all(isinstance(val, float) for val in result.values())
        assert all(0 <= val <= 1.0 for val in result.values())

    def test_doc(self, spacy_doc):
        result = RESOURCE.get_emotional_valence(spacy_doc)
        assert isinstance(result, dict)
        assert all(isinstance(key, str) for key in result.keys())
        assert all(isinstance(val, float) for val in result.values())
        assert all(0 <= val <= 1.0 for val in result.values())



import pytest
from spacy.tokens import Doc

from textacy import load_spacy_lang
from textacy.spacier import utils


@pytest.fixture(scope="module")
def spacy_doc():
    spacy_lang = load_spacy_lang("en_core_web_sm")
    text = (
        "The unit tests aren't going well. "
        "I love Python, but I don't love backwards incompatibilities. "
        "No programmers were permanently damaged for textacy's sake. "
        "Thank God for Stack Overflow."
    )
    spacy_doc = spacy_lang(text.strip())
    return spacy_doc


def test_preserve_case(spacy_doc):
    results = [utils.preserve_case(tok) for tok in spacy_doc]
    assert all(isinstance(result, bool) for result in results)
    assert (
        sum(1 if result else 0 for result in results) <
        sum(1 if not result else 0 for result in results)
    )


def test_get_normalized_text(spacy_doc):
    expected = [
        "the",
        "unit",
        "test",
        "be",
        "n't",
        "go",
        "well",
        ".",
        "I",
        "love",
        "Python",
        ",",
        "but",
        "I",
        "do",
        "n't",
        "love",
        "backwards",
        "incompatibility",
        ".",
        "no",
        "programmer",
        "be",
        "permanently",
        "damage",
        "for",
        "textacy",
        "'s",
        "sake",
        ".",
        "thank",
        "God",
        "for",
        "Stack",
        "Overflow",
        ".",
    ]
    observed = [utils.get_normalized_text(tok) for tok in spacy_doc if not tok.is_space]
    assert observed == expected


def test_get_main_verbs_of_sent(spacy_doc):
    expected = [["going"], ["love", "love"], ["damaged"], ["Thank"]]
    observed = [
        [tok.text for tok in utils.get_main_verbs_of_sent(sent)]
        for sent in spacy_doc.sents
    ]
    for obs, exp in zip(observed, expected):
        assert obs == exp


def test_get_subjects_of_verb(spacy_doc):
    expected = [["tests"], ["I"], ["I"], ["programmers"], []]
    main_verbs = [
        tok for sent in spacy_doc.sents for tok in utils.get_main_verbs_of_sent(sent)
    ]
    observed = [
        [tok.text for tok in utils.get_subjects_of_verb(main_verb)]
        for main_verb in main_verbs
    ]
    for obs, exp in zip(observed, expected):
        assert obs == exp


def test_get_objects_of_verb(spacy_doc):
    expected = [[], ["Python"], ["incompatibilities"], [], ["God"]]
    main_verbs = [
        tok for sent in spacy_doc.sents for tok in utils.get_main_verbs_of_sent(sent)
    ]
    observed = [
        [tok.text for tok in utils.get_objects_of_verb(main_verb)]
        for main_verb in main_verbs
    ]
    for obs, exp in zip(observed, expected):
        assert obs == exp


def test_make_doc_from_text_chunks():
    text = "Burton forgot to add tests for this function."
    for lang in ("en_core_web_sm", load_spacy_lang("en_core_web_sm")):
        spacy_doc = utils.make_doc_from_text_chunks(text, lang)
        assert isinstance(spacy_doc, Doc)
        assert spacy_doc.text == text



"""
Networks
--------

:mod:`textacy.representations.network`: Represent document data as networks,
where nodes are terms, sentences, or even full documents and edges between them
are weighted by the strength of their co-occurrence or similarity.
"""
from __future__ import annotations

import collections
import itertools
import logging
from operator import itemgetter
from typing import Any, Collection, Dict, Optional, Sequence, Set, Union

import networkx as nx
import numpy as np
from cytoolz import itertoolz

from .. import errors, similarity


LOGGER = logging.getLogger(__name__)


def build_cooccurrence_network(
    data: Sequence[str] | Sequence[Sequence[str]],
    *,
    window_size: int = 2,
    edge_weighting: str = "count",  # Literal["count", "binary"]
) -> nx.Graph:
    """
    Transform an ordered sequence of strings (or a sequence of such sequences)
    into a graph, where each string is represented by a node with weighted edges
    linking it to other strings that co-occur within ``window_size`` elements of itself.

    Input ``data`` can take a variety of forms. For example, as a ``Sequence[str]``
    where elements are token or term strings from a single document:

    .. code-block:: pycon

        >>> texts = [
        ...     "Mary had a little lamb. Its fleece was white as snow.",
        ...     "Everywhere that Mary went the lamb was sure to go.",
        ... ]
        >>> docs = [make_spacy_doc(text, lang="en_core_web_sm") for text in texts]
        >>> data = [tok.text for tok in docs[0]]
        >>> graph = build_cooccurrence_network(data, window_size=2)
        >>> sorted(graph.adjacency())[0]
        ('.', {'lamb': {'weight': 1}, 'Its': {'weight': 1}, 'snow': {'weight': 1}})

    Or as a ``Sequence[Sequence[str]]``, where elements are token or term strings
    per sentence from a single document:

    .. code-block:: pycon

        >>> data = [[tok.text for tok in sent] for sent in docs[0].sents]
        >>> graph = build_cooccurrence_network(data, window_size=2)
        >>> sorted(graph.adjacency())[0]
        ('.', {'lamb': {'weight': 1}, 'snow': {'weight': 1}})

    Or as a ``Sequence[Sequence[str]]``, where elements are token or term strings
    per document from multiple documents:

    .. code-block:: pycon

        >>> data = [[tok.text for tok in doc] for doc in docs]
        >>> graph = build_cooccurrence_network(data, window_size=2)
        >>> sorted(graph.adjacency())[0]
        ('.',
         {'lamb': {'weight': 1},
          'Its': {'weight': 1},
          'snow': {'weight': 1},
          'go': {'weight': 1}})

    Note how the "." token's connections to other nodes change for each case. (Note
    that in real usage, you'll probably want to remove stopwords, punctuation, etc.
    so that nodes in the graph represent meaningful concepts.)

    Args:
        data
        window_size: Size of sliding window over ``data`` that determines
            which strings are said to co-occur. For example, a value of 2 means that
            only immediately adjacent strings will have edges in the network;
            larger values loosen the definition of co-occurrence and typically
            lead to a more densely-connected network.

            .. note:: Co-occurrence windows are not permitted to cross sequences.
               So, if ``data`` is a ``Sequence[Sequence[str]]``, then co-occ counts
               are computed separately for each sub-sequence, then summed together.

        edge_weighting: Method by which edges between nodes are weighted.
            If "count", nodes are connected by edges with weights equal to
            the number of times they co-occurred within a sliding window;
            if "binary", all such edges have weight set equal to 1.

    Returns:
        Graph whose nodes correspond to individual strings from ``data``;
        those that co-occur are connected by edges with weights determined
        by ``edge_weighting``.

    Reference:
        https://en.wikipedia.org/wiki/Co-occurrence_network
    """
    if not data:
        LOGGER.warning("input `data` is empty, so output graph is also empty")
        return nx.Graph()

    if window_size < 2:
        raise ValueError(f"window_size = {window_size} is invalid; value must be >= 2")

    # input data is Sequence[str]
    if isinstance(data[0], str):
        windows = itertoolz.sliding_window(min(window_size, len(data)), data)
    # input data is Sequence[Sequence[str]]
    elif isinstance(data[0], Sequence) and isinstance(data[0][0], str):
        windows = itertoolz.concat(
            itertoolz.sliding_window(min(window_size, len(subseq)), subseq)
            for subseq in data
        )
    else:
        raise TypeError(
            errors.type_invalid_msg(
                "data", data, Union[Sequence[str], Sequence[Sequence[str]]]
            )
        )

    graph = nx.Graph()
    if edge_weighting == "count":
        cooc_counts = collections.Counter(
            w1_w2
            for window in windows
            for w1_w2 in itertools.combinations(sorted(window), 2)
        )
        graph.add_edges_from(
            (w1, w2, {"weight": weight}) for (w1, w2), weight in cooc_counts.items()
        )
    elif edge_weighting == "binary":
        edge_data = {"weight": 1}
        graph.add_edges_from(
            (w1, w2, edge_data)
            for window in windows for (w1, w2) in itertools.combinations(window, 2)
        )
    else:
        raise ValueError(
            errors.value_invalid_msg(
                "edge_weighting", edge_weighting, {"count", "binary"}
            )
        )

    return graph


def build_similarity_network(
    data: Sequence[str] | Sequence[Sequence[str]], edge_weighting: str,
) -> nx.Graph:
    """
    Transform a sequence of strings (or a sequence of such sequences) into a graph,
    where each element of the top-level sequence is represented by a node
    with edges linking it to all other elements weighted by their pairwise similarity.

    Input ``data`` can take a variety of forms. For example, as a ``Sequence[str]``
    where elements are sentence texts from a single document:

    .. code-block:: pycon

        >>> texts = [
        ...     "Mary had a little lamb. Its fleece was white as snow.",
        ...     "Everywhere that Mary went the lamb was sure to go.",
        ... ]
        >>> docs = [make_spacy_doc(text, lang="en_core_web_sm") for text in texts]
        >>> data = [sent.text.lower() for sent in docs[0].sents]
        >>> graph = build_similarity_network(data, "levenshtein")
        >>> sorted(graph.adjacency())[0]
        ('its fleece was white as snow.',
         {'mary had a little lamb.': {'weight': 0.24137931034482762}})

    Or as a ``Sequence[str]`` where elements are full texts from multiple documents:

    .. code-block:: pycon

        >>> data = [doc.text.lower() for doc in docs]
        >>> graph = build_similarity_network(data, "jaro")
        >>> sorted(graph.adjacency())[0]
        ('everywhere that mary went the lamb was sure to go.',
         {'mary had a little lamb. its fleece was white as snow.': {'weight': 0.6516002795248078}})

    Or as a ``Sequence[Sequence[str]]`` where elements are tokenized texts from
    multiple documents:

    .. code-block:: pycon

        >>> data = [[tok.lower_ for tok in doc] for doc in docs]
        >>> graph = build_similarity_network(data, "jaccard")
        >>> sorted(graph.adjacency())[0]
        (('everywhere', 'that', 'mary', 'went', 'the', 'lamb', 'was', 'sure', 'to', 'go', '.'),
         {('mary', 'had', 'a', 'little', 'lamb', '.', 'its', 'fleece', 'was', 'white', 'as', 'snow', '.'): {'weight': 0.21052631578947367}})

    Args:
        data
        edge_weighting: Similarity metric to use for weighting edges between elements
            in ``data``, represented as the name of a function available in
            :mod:`textacy.similarity`.

            .. note:: Different metrics are suited for different forms and contexts
               of ``data``. You'll have to decide which method makes sense. For example,
               when comparing a sequence of short strings, "levenshtein" is often
               a reasonable bet; when comparing a sequence of sequences of somewhat
               noisy strings (e.g. includes punctuation, cruft tokens), you might try
               "matching_subsequences_ratio" to help filter out the noise.

    Returns:
        Graph whose nodes correspond to top-level sequence elements in ``data``,
        connected by edges to all other nodes with weights determined by
        their pairwise similarity.

    Reference:
        https://en.wikipedia.org/wiki/Semantic_similarity_network -- this is *not*
        the same as what's implemented here, but they're similar in spirit.
    """
    if not data:
        LOGGER.warning("input `data` is empty, so output graph is also empty")
        return nx.Graph()

    sim_func = getattr(similarity, edge_weighting)

    if isinstance(data[0], str):
        ele_pairs = itertools.combinations(data, 2)
    elif isinstance(data[0], Sequence) and isinstance(data[0][0], str):
        # nx graph nodes need to be *hashable*, so Sequence => Tuple
        ele_pairs = (
            (tuple(ele1), tuple(ele2))
            for ele1, ele2 in itertools.combinations(data, 2)
        )
    else:
        raise TypeError(
            errors.type_invalid_msg(
                "data", data, Union[Sequence[str], Sequence[Sequence[str]]]
            )
        )

    graph = nx.Graph()
    graph.add_edges_from(
        (ele1, ele2, {"weight": sim_func(ele1, ele2)})
        for ele1, ele2 in ele_pairs
    )
    return graph


def rank_nodes_by_pagerank(
    graph: nx.Graph, weight: str = "weight", **kwargs,
) -> Dict[Any, float]:
    """
    Rank nodes in ``graph`` using the Pagegrank algorithm.

    Args:
        graph
        weight: Key in edge data that holds weights.
        **kwargs

    Returns:
        Mapping of node object to Pagerank score.
    """
    return nx.pagerank_scipy(graph, weight=weight, **kwargs)


def rank_nodes_by_bestcoverage(
    graph: nx.Graph, k: int, c: int = 1, alpha: float = 1.0, weight: str = "weight",
) -> Dict[Any, float]:
    """
    Rank nodes in a network using the BestCoverage algorithm that attempts to
    balance between node centrality and diversity.

    Args:
        graph
        k: Number of results to return for top-k search.
        c: *l* parameter for *l*-step expansion; best if 1 or 2
        alpha: Float in [0.0, 1.0] specifying how much of central vertex's score
            to remove from its *l*-step neighbors; smaller value puts more emphasis
            on centrality, larger value puts more emphasis on diversity
        weight: Key in edge data that holds weights.

    Returns:
        Top ``k`` nodes as ranked by bestcoverage algorithm; keys as node
        identifiers, values as corresponding ranking scores

    References:
        Küçüktunç, O., Saule, E., Kaya, K., & Çatalyürek, Ü. V. (2013, May).
        Diversified recommendation on graphs: pitfalls, measures, and algorithms.
        In Proceedings of the 22nd international conference on World Wide Web
        (pp. 715-726). International World Wide Web Conferences Steering Committee.
        http://www2013.wwwconference.org/proceedings/p715.pdf
    """
    alpha = float(alpha)

    nodes_list = [node for node in graph]
    if len(nodes_list) == 0:
        LOGGER.warning("`graph` is empty")
        return {}

    # ranks: array of PageRank values, summing up to 1
    ranks = nx.pagerank_scipy(graph, alpha=0.85, max_iter=100, tol=1e-08, weight=weight)
    # sorted_ranks = sorted(ranks.items(), key=itemgetter(1), reverse=True)
    # avg_degree = sum(dict(graph.degree()).values()) / len(nodes_list)
    # relaxation parameter, k' in the paper
    # k_prime = int(k * avg_degree * c)

    # top_k_sorted_ranks = sorted_ranks[:k_prime]

    def get_l_step_expanded_set(vertices: Collection[str], n_steps: int) -> Set[str]:
        """
        Args:
            vertices: vertices to be expanded
            n_steps: how many steps to expand vertices set

        Returns:
            the l-step expanded set of vertices
        """
        # add vertices to s
        s = set(vertices)
        # s.update(vertices)
        # for each step
        for _ in range(n_steps):
            # for each node
            next_vertices = []
            for vertex in vertices:
                # add its neighbors to the next list
                neighbors = graph.neighbors(vertex)
                next_vertices.extend(neighbors)
                s.update(neighbors)
            vertices = set(next_vertices)
        return s

    # TODO: someday, burton, figure out what you were going to do with this...
    # top_k_exp_vertices = get_l_step_expanded_set(
    #     [item[0] for item in top_k_sorted_ranks], c
    # )

    # compute initial exprel contribution
    taken = collections.defaultdict(bool)
    contrib = {}
    for vertex in nodes_list:
        # get l-step expanded set
        s = get_l_step_expanded_set([vertex], c)
        # sum up neighbors ranks, i.e. l-step expanded relevance
        contrib[vertex] = sum(ranks[v] for v in s)

    sum_contrib = 0.0
    results = {}
    # greedily select to maximize exprel metric
    for _ in range(k):
        if not contrib:
            break
        # find word with highest l-step expanded relevance score
        max_word_score = sorted(contrib.items(), key=itemgetter(1), reverse=True)[0]
        sum_contrib += max_word_score[1]  # contrib[max_word[0]]
        results[max_word_score[0]] = max_word_score[1]
        # find its l-step expanded set
        l_step_expanded_set = get_l_step_expanded_set([max_word_score[0]], c)
        # for each vertex found
        for vertex in l_step_expanded_set:
            # already removed its contribution from neighbors
            if taken[vertex] is True:
                continue
            # remove the contribution of vertex (or some fraction) from its l-step neighbors
            s1 = get_l_step_expanded_set([vertex], c)
            for w in s1:
                try:
                    contrib[w] -= alpha * ranks[vertex]
                except KeyError:
                    LOGGER.error(
                        "Word %s not in contrib dict! We're approximating...", w
                    )
            taken[vertex] = True
        contrib[max_word_score[0]] = 0

    return results


def rank_nodes_by_divrank(
    graph: nx.Graph,
    r: Optional[np.ndarray] = None,
    lambda_: float = 0.5,
    alpha: float = 0.5,
) -> Dict[str, float]:
    """
    Rank nodes in a network using the DivRank algorithm that attempts to
    balance between node centrality and diversity.

    Args:
        graph
        r: The "personalization vector"; by default, ``r = ones(1, n)/n``
        lambda_: Float in [0.0, 1.0]
        alpha: Float in [0.0, 1.0] that controls the strength of self-links.

    Returns:
        Mapping of node to score ordered by descending divrank score

    References:
        Mei, Q., Guo, J., & Radev, D. (2010, July). Divrank: the interplay of
        prestige and diversity in information networks. In Proceedings of the
        16th ACM SIGKDD international conference on Knowledge discovery and data
        mining (pp. 1009-1018). ACM.
        http://clair.si.umich.edu/~radev/papers/SIGKDD2010.pdf
    """
    # check function arguments
    if len(graph) == 0:
        LOGGER.warning("`graph` is empty")
        return {}

    # specify the order of nodes to use in creating the matrix
    # and then later recovering the values from the order index
    nodes_list = [node for node in graph]
    # create adjacency matrix, i.e.
    # n x n matrix where entry W_ij is the weight of the edge from V_i to V_j
    W = nx.to_numpy_matrix(graph, nodelist=nodes_list, weight="weight").A
    n = W.shape[1]
    # create flat prior personalization vector if none given
    if r is None:
        r = np.array([n * [1 / float(n)]])
    # Specify some constants
    max_iter = 1000
    diff = 1e10
    tol = 1e-3
    pr = np.array([n * [1 / float(n)]])
    # Get p0(v -> u), i.e. transition probability prior to reinforcement
    tmp = np.reshape(np.sum(W, axis=1), (n, 1))
    idx_nan = np.flatnonzero(tmp == 0)
    W0 = W / np.tile(tmp, (1, n))
    W0[idx_nan, :] = 0
    del W

    # DivRank algorithm
    i = 0
    while i < max_iter and diff > tol:
        W1 = alpha * W0 * np.tile(pr, (n, 1))
        W1 = W1 - np.diag(W1[:, 0]) + (1 - alpha) * np.diag(pr[0, :])
        tmp1 = np.reshape(np.sum(W1, axis=1), (n, 1))
        P = W1 / np.tile(tmp1, (1, n))
        P = ((1 - lambda_) * P) + (lambda_ * np.tile(r, (n, 1)))
        pr_new = np.dot(pr, P)
        i += 1
        diff = np.sum(np.abs(pr_new - pr)) / np.sum(pr)
        pr = pr_new

    # sort nodes by divrank score
    results = sorted(
        ((i, score) for i, score in enumerate(pr.flatten().tolist())),
        key=itemgetter(1),
        reverse=True,
    )

    # replace node number by node value
    divranks = {nodes_list[result[0]]: result[1] for result in results}

    return divranks



import pytest

from textacy import preprocessing


@pytest.mark.parametrize(
    "text_in, text_out",
    [
        ("• foo\n• bar", "- foo\n- bar"),
        ("• foo\n    • bar", "- foo\n    - bar"),
        (
            "\n‣ item1\n⁃ item2\n⁌ item3\n⁍ item4\n∙ item5\n▪ item6\n● item7\n◦ item8",
            "\n- item1\n- item2\n- item3\n- item4\n- item5\n- item6\n- item7\n- item8",
        ),
        (
            "\n⦾ item1\n⦿ item2\n・ item3",
            "\n- item1\n- item2\n- item3",
        ),
    ]
)
def test_normalize_bullet_points(text_in, text_out):
    assert preprocessing.normalize.bullet_points(text_in) == text_out


@pytest.mark.parametrize(
    "text_in, text_out",
    [
        ("I see you shiver with antici- pation.", "I see you shiver with anticipation."),
        ("I see you shiver with antici-   \npation.", "I see you shiver with anticipation."),
        ("I see you shiver with antici- PATION.", "I see you shiver with anticiPATION."),
        ("I see you shiver with antici- 1pation.", "I see you shiver with antici- 1pation."),
        ("I see you shiver with antici pation.", "I see you shiver with antici pation."),
        ("I see you shiver with antici-pation.", "I see you shiver with antici-pation."),
        ("My phone number is 555- 1234.", "My phone number is 555- 1234."),
        ("I got an A- on the test.", "I got an A- on the test."),
    ]
)
def test_normalize_hyphenated_words(text_in, text_out):
    assert preprocessing.normalize.hyphenated_words(text_in) == text_out


@pytest.mark.parametrize(
    "text_in, text_out",
    [
        ("These are ´funny single quotes´.", "These are 'funny single quotes'."),
        ("These are ‘fancy single quotes’.", "These are 'fancy single quotes'."),
        ("These are “fancy double quotes”.", "These are \"fancy double quotes\"."),
    ]
)
def test_normalize_quotation_marks(text_in, text_out):
    assert preprocessing.normalize.quotation_marks(text_in) == text_out


@pytest.mark.parametrize(
    "kwargs, text_out",
    [
        (
            dict(chars=".", maxn=3),
            "**Hello**, world!!! I wonder... How are *you* doing?!?! lololol",
        ),
        (
            dict(chars="*", maxn=1),
            "*Hello*, world!!! I wonder....... How are *you* doing?!?! lololol",
        ),
        (
            dict(chars="?!", maxn=1),
            "**Hello**, world!!! I wonder....... How are *you* doing?! lololol",
        ),
        (
            dict(chars="ol", maxn=2),
            "**Hello**, world!!! I wonder....... How are *you* doing?!?! lolol",
        ),
        (
            dict(chars="*", maxn=0),
            "Hello, world!!! I wonder....... How are you doing?!?! lololol",
        ),
    ]
)
def test_normalize_repeating_chars(kwargs, text_out):
    text_in = "**Hello**, world!!! I wonder....... How are *you* doing?!?! lololol"
    assert preprocessing.normalize.repeating_chars(text_in, **kwargs) == text_out


@pytest.mark.parametrize(
    "text_in, text_out",
    [
        ("Well… That's a long story.", "Well... That's a long story."),
    ]
)
def test_normalize_unicode(text_in, text_out):
    assert preprocessing.normalize.unicode(text_in, form="NFKC") == text_out


@pytest.mark.parametrize(
    "text_in, text_out",
    [
        ("Hello,  world!", "Hello, world!"),
        ("Hello,     world!", "Hello, world!"),
        ("Hello,\tworld!", "Hello, world!"),
        ("Hello,\t\t  world!", "Hello, world!"),
        ("Hello,\n\nworld!", "Hello,\nworld!"),
        ("Hello,\r\nworld!", "Hello,\nworld!"),
        ("Hello\uFEFF, world!", "Hello, world!"),
        ("Hello\u200B\u200B, world!", "Hello, world!"),
        ("Hello\uFEFF,\n\n\nworld   !  ", "Hello,\nworld !"),
    ]
)
def test_normalize_whitespace(text_in, text_out):
    assert preprocessing.normalize.whitespace(text_in) == text_out



import pytest

from textacy.similarity import tokens


@pytest.mark.parametrize(
    "seq1, seq2, exp",
    [
        (list("abcd"), list("abcd"), 1.0),
        (list("abcd"), list("wxyz"), 0.0),
        (list("abcd"), [], 0.0),
        ([], [], 0.0),
        (list("abcd"), list("abcD"), 0.6),
        (list("Abcd"), list("abcd"), 0.6),
        (list("abcd"), list("abcdefgh"), 0.5),
    ]
)
def test_jaccard(seq1, seq2, exp):
    obs = tokens.jaccard(seq1, seq2)
    assert isinstance(obs, float)
    assert 0.0 <= obs <= 1.0
    assert obs == pytest.approx(exp, rel=0.01)


@pytest.mark.parametrize(
    "seq1, seq2, exp",
    [
        (list("abcd"), list("abcd"), 1.0),
        (list("abcd"), list("wxyz"), 0.0),
        (list("abcd"), [], 0.0),
        ([], [], 0.0),
        (list("abcd"), list("abcD"), 0.75),
        (list("Abcd"), list("abcd"), 0.75),
        (list("abcd"), list("abcdefgh"), 0.6666),
    ]
)
def test_sorensen_dice(seq1, seq2, exp):
    obs = tokens.sorensen_dice(seq1, seq2)
    assert isinstance(obs, float)
    assert 0.0 <= obs <= 1.0
    assert obs == pytest.approx(exp, rel=0.01)


@pytest.mark.parametrize(
    "seq1, seq2, alpha, beta, exp",
    [
        (list("abcd"), list("abcd"), 1.0, 1.0, 1.0),
        (list("abcd"), list("wxyz"), 1.0, 1.0, 0.0),
        (list("abcd"), [], 1.0, 1.0, 0.0),
        ([], [], 1.0, 1.0, 0.0),
        (list("abcd"), list("abcD"), 1.0, 1.0, 0.75),
        (list("Abcd"), list("abcd"), 1.0, 1.0, 0.75),
        (list("abcd"), list("abcdefgh"), 1.0, 1.0, 1.0),
        (list("abcd"), list("abcdefgh"), 0.5, 1.0, 0.6666),
        (list("abcd"), list("abcdefgh"), 0.5, 2.0, 0.5),
    ]
)
def test_tversky(seq1, seq2, alpha, beta, exp):
    obs = tokens.tversky(seq1, seq2, alpha=alpha, beta=beta)
    assert isinstance(obs, float)
    assert 0.0 <= obs <= 1.0
    assert obs == pytest.approx(exp, rel=0.01)


@pytest.mark.parametrize(
    "seq1, seq2, alpha, beta, equiv",
    [
        (list("abcd"), list("abcD"), 0.5, 2.0, tokens.jaccard),
        (list("abcd"), list("abcdefgh"), 0.5, 2.0, tokens.jaccard),
        (list("abcd"), list("abcD"), 0.5, 1.0, tokens.sorensen_dice),
        (list("abcd"), list("abcdefgh"), 0.5, 1.0, tokens.sorensen_dice),
    ]
)
def test_tversky_equivalence(seq1, seq2, alpha, beta, equiv):
    obs = tokens.tversky(seq1, seq2, alpha=alpha, beta=beta)
    obs_equiv = equiv(seq1, seq2)
    assert obs == obs_equiv


@pytest.mark.parametrize(
    "seq1, seq2, exp",
    [
        (list("abcd"), list("abcd"), 1.0),
        (list("abcd"), list("wxyz"), 0.0),
        (list("abcd"), [], 0.0),
        ([], [], 0.0),
        (list("abcd"), list("abcD"), 0.75),
        (list("Abcd"), list("abcd"), 0.75),
        (list("abcd"), list("abcdefgh"), 0.7071),
    ]
)
def test_cosine(seq1, seq2, exp):
    obs = tokens.cosine(seq1, seq2)
    assert isinstance(obs, float)
    assert 0.0 <= obs <= 1.0
    assert obs == pytest.approx(exp, rel=0.01)


@pytest.mark.parametrize(
    "seq1, seq2, exp",
    [
        (list("abcd"), list("abcd"), 1.0),
        (list("abcd"), list("wxyz"), 0.0),
        (list("abcd"), [], 0.0),
        ([], [], 0.0),
        (list("abcd"), list("abcD"), 0.75),
        (list("Abcd"), list("abcd"), 0.75),
        (list("abcd"), list("abcdefgh"), 0.5),
        (list("abcd"), list("aabcd"), 0.8),
        (list("abcd"), list("aabcdd"), 0.6666),
        (list("abcd"), list("aabbccdd"), 0.5),
    ]
)
def test_bag(seq1, seq2, exp):
    obs = tokens.bag(seq1, seq2)
    assert isinstance(obs, float)
    assert 0.0 <= obs <= 1.0
    assert obs == pytest.approx(exp, rel=0.01)



"""
spaCy Utils
-----------

:mod:`textacy.spacier.utils`: Helper functions for working with / extending spaCy's
core functionality.
"""
from __future__ import annotations

import itertools
import pathlib
from typing import Iterable, List, Tuple, Union

from spacy import attrs
from spacy.language import Language
from spacy.symbols import PROPN, VERB
from spacy.tokens import Doc, Span, Token

from .. import constants, errors, types, utils
from . import core


def make_doc_from_text_chunks(
    text: str, lang: types.LangLike, chunk_size: int = 100000,
) -> Doc:
    """
    Make a single spaCy-processed document from 1 or more chunks of ``text``.
    This is a workaround for processing very long texts, for which spaCy
    is unable to allocate enough RAM.

    Args:
        text: Text document to be chunked and processed by spaCy.
        lang: Language with which spaCy processes ``text``, represented as
            the full name of or path on disk to the pipeline, or
            an already instantiated pipeline instance.
        chunk_size: Number of characters comprising each text chunk
            (excluding the last chunk, which is probably smaller).
            For best performance, value should be somewhere between 1e3 and 1e7,
            depending on how much RAM you have available.

            .. note:: Since chunking is done by character, chunks edges' probably
               won't respect natural language segmentation, which means that every
               ``chunk_size`` characters, spaCy's models may make mistakes.

    Returns:
        A single processed document, built from concatenated text chunks.
    """
    utils.deprecated(
        "This function is deprecated, and will be removed in a future version. "
        "Instead, use the usual :func:`textacy.make_spacy_doc()` "
        "and specify a non-null `chunk_size`",
        action="once",
    )
    lang = resolve_langlike(lang)
    text_chunks = (text[i : i + chunk_size] for i in range(0, len(text), chunk_size))
    docs = list(lang.pipe(text_chunks))
    return Doc.from_docs(docs)


def merge_spans(spans: Iterable[Span], doc: Doc) -> None:
    """
    Merge spans into single tokens in ``doc``, *in-place*.

    Args:
        spans (Iterable[:class:`spacy.tokens.Span`])
        doc (:class:`spacy.tokens.Doc`)
    """
    with doc.retokenize() as retokenizer:
        string_store = doc.vocab.strings
        for span in spans:
            retokenizer.merge(
                doc[span.start : span.end],
                attrs=attrs.intify_attrs({"ent_type": span.label}, string_store),
            )


def preserve_case(token: Token) -> bool:
    """
    Return True if ``token`` is a proper noun or acronym; otherwise, False.

    Raises:
        ValueError: If parent document has not been POS-tagged.
    """
    if token.doc.has_annotation("TAG") is False:
        raise ValueError(f"parent doc of token '{token}' has not been POS-tagged")
    # is_acronym() got moved into a subpkg with heavier dependencies that we don't
    # want imported at the top-level package; this is the only outside place
    # that uses that function, so let's hide the required import in this function
    # using a try/except for performance
    # not the prettiest solution, but should be alright
    try:
        acros.is_acronym
    except NameError:
        from textacy.extract import acros
    if token.pos == PROPN or acros.is_acronym(token.text):
        return True
    else:
        return False


def get_normalized_text(span_or_token: Span | Token) -> str:
    """
    Get the text of a spaCy span or token, normalized depending on its characteristics.
    For proper nouns and acronyms, text is returned as-is; for everything else,
    text is lemmatized.
    """
    if isinstance(span_or_token, Token):
        return (
            span_or_token.text if preserve_case(span_or_token) else span_or_token.lemma_
        )
    elif isinstance(span_or_token, Span):
        return " ".join(
            token.text if preserve_case(token) else token.lemma_
            for token in span_or_token
        )
    else:
        raise TypeError(
            errors.type_invalid_msg(
                "span_or_token", type(span_or_token), Union[Span, Token]
            )
        )


def get_main_verbs_of_sent(sent: Span) -> List[Token]:
    """Return the main (non-auxiliary) verbs in a sentence."""
    return [
        tok for tok in sent if tok.pos == VERB and tok.dep_ not in constants.AUX_DEPS
    ]


def get_subjects_of_verb(verb: Token) -> List[Token]:
    """Return all subjects of a verb according to the dependency parse."""
    subjs = [tok for tok in verb.lefts if tok.dep_ in constants.SUBJ_DEPS]
    # get additional conjunct subjects
    subjs.extend(tok for subj in subjs for tok in _get_conjuncts(subj))
    return subjs


def get_objects_of_verb(verb: Token) -> List[Token]:
    """
    Return all objects of a verb according to the dependency parse,
    including open clausal complements.
    """
    objs = [tok for tok in verb.rights if tok.dep_ in constants.OBJ_DEPS]
    # get open clausal complements (xcomp)
    objs.extend(tok for tok in verb.rights if tok.dep_ == "xcomp")
    # get additional conjunct objects
    objs.extend(tok for obj in objs for tok in _get_conjuncts(obj))
    return objs


def _get_conjuncts(tok: Token) -> List[Token]:
    """
    Return conjunct dependents of the leftmost conjunct in a coordinated phrase,
    e.g. "Burton, [Dan], and [Josh] ...".
    """
    return [right for right in tok.rights if right.dep_ == "conj"]


def get_span_for_compound_noun(noun: Token) -> Tuple[int, int]:
    """Return document indexes spanning all (adjacent) tokens in a compound noun."""
    min_i = noun.i - sum(
        1
        for _ in itertools.takewhile(
            lambda x: x.dep_ == "compound", reversed(list(noun.lefts))
        )
    )
    return (min_i, noun.i)


def get_span_for_verb_auxiliaries(verb: Token) -> Tuple[int, int]:
    """
    Return document indexes spanning all (adjacent) tokens
    around a verb that are auxiliary verbs or negations.
    """
    min_i = verb.i - sum(
        1
        for _ in itertools.takewhile(
            lambda x: x.dep_ in constants.AUX_DEPS, reversed(list(verb.lefts))
        )
    )
    max_i = verb.i + sum(
        1
        for _ in itertools.takewhile(lambda x: x.dep_ in constants.AUX_DEPS, verb.rights)
    )
    return (min_i, max_i)


def resolve_langlike(lang: types.LangLike) -> Language:
    if isinstance(lang, Language):
        return lang
    elif isinstance(lang, (str, pathlib.Path)):
        return core.load_spacy_lang(lang)
    else:
        raise TypeError(errors.type_invalid_msg("lang", type(lang), types.LangLike))


def resolve_langlikeincontext(text: str, lang: types.LangLikeInContext) -> Language:
    if isinstance(lang, Language):
        return lang
    elif isinstance(lang, (str, pathlib.Path)):
        return core.load_spacy_lang(lang)
    elif callable(lang):
        return resolve_langlikeincontext(text, lang(text))
    else:
        raise TypeError(
            errors.type_invalid_msg("lang", type(lang), types.LangLikeInContext)
        )



import pytest

from textacy import load_spacy_lang
from textacy.text_stats import components

TEXT = (
    "The year was 2081, and everybody was finally equal. "
    "They weren't only equal before God and the law. "
    "They were equal every which way."
)


@pytest.fixture(scope="module")
def spacy_lang():
    spacy_lang = load_spacy_lang("en_core_web_sm")
    spacy_lang.add_pipe("textacy_text_stats", last=True)

    yield spacy_lang

    # remove component after running these tests
    spacy_lang.remove_pipe("textacy_text_stats")


@pytest.fixture(scope="module")
def spacy_doc(spacy_lang):
    spacy_doc = spacy_lang(TEXT)
    return spacy_doc


def test_component_name(spacy_lang):
    assert spacy_lang.has_pipe("textacy_text_stats") is True


def test_component_problems(spacy_lang):
    assert spacy_lang.analyze_pipes()["problems"]["textacy_text_stats"] == []


@pytest.mark.parametrize(
    "attrs",
    [
        None,
        "flesch_kincaid_grade_level",
        ["flesch_kincaid_grade_level", "flesch_reading_ease"],
    ],
)
def test_component_attrs(attrs):
    text_stats_component = components.TextStatsComponent(attrs=attrs)
    assert isinstance(text_stats_component.attrs, tuple) is True


def test_attrs_on_doc(spacy_lang, spacy_doc):
    tsc = spacy_lang.get_pipe("textacy_text_stats")
    for attr in tsc.attrs:
        assert spacy_doc._.has(attr) is True
        assert isinstance(spacy_doc._.get(attr), (int, float, tuple)) is True



import re

import pytest

import textacy
from textacy import extract


@pytest.fixture(scope="module")
def text():
    return (
        "Today you are You! That is truer than true! "
        "There is no one alive who is You-er than you!"
    )


@pytest.fixture(scope="module")
def doc(text):
    nlp = textacy.load_spacy_lang("en_core_web_sm")
    return nlp(text)


class TestKeywordInContext:

    def test_doc_type(self, text, doc):
        text_result = list(extract.keyword_in_context(text, "you"))
        doc_result = list(extract.keyword_in_context(doc, "you"))
        assert text_result == doc_result

    @pytest.mark.parametrize(
        "keyword, ignore_case",
        [
            ("you", True),
            ("You", False),
            ("you(-er)?", True),
            # ignore_case shouldn't affect anything when keyword is compiled regex
            (re.compile("you", flags=re.IGNORECASE), False),
            (re.compile("you(-er)", flags=re.IGNORECASE), True),
        ]
    )
    def test_keywords(self, text, keyword, ignore_case):
        obs = list(
            extract.keyword_in_context(
                text,
                keyword,
                ignore_case=ignore_case,
            )
        )
        if isinstance(keyword, str):
            assert all(
                re.match(keyword, kw, flags=0 if ignore_case is False else re.IGNORECASE)
                for _, kw, _ in obs
            )
        else:
            assert all(re.match(keyword, kw) for _, kw, _ in obs)

    @pytest.mark.parametrize(
        "window_width, pad_context",
        [
            (10, False),
            (10, True),
            (3, True),
            (3, False),
        ]
    )
    def test_contexts(self, text, window_width, pad_context):
        obs = list(
            extract.keyword_in_context(
                text,
                "you",
                window_width=window_width,
                pad_context=pad_context,
            )
        )
        if pad_context is False:
            assert all(
                len(pre_ctx) <= window_width and len(post_ctx) <= window_width
                for pre_ctx, _, post_ctx in obs
            )
        else:
            assert all(
                len(pre_ctx) == window_width and len(post_ctx) == window_width
                for pre_ctx, _, post_ctx in obs
            )

    @pytest.mark.parametrize(
        "keyword, ignore_case, window_width, pad_context, exp",
        [
            (
                "you",
                True,
                10,
                False,
                [
                    ('Today ', 'you', ' are You! '),
                    ('y you are ', 'You', '! That is '),
                    ('ve who is ', 'You', '-er than y'),
                    ('u-er than ', 'you', '!'),
                ],
            ),
            (
                "you",
                False,
                10,
                False,
                [
                    ('Today ', 'you', ' are You! '),
                    ('u-er than ', 'you', '!'),
                ],
            ),
            (
                "you",
                True,
                10,
                True,
                [
                    ('    Today ', 'you', ' are You! '),
                    ('y you are ', 'You', '! That is '),
                    ('ve who is ', 'You', '-er than y'),
                    ('u-er than ', 'you', '!         '),
                ],
            ),
            (
                re.compile("you", flags=re.IGNORECASE),
                True,
                10,
                False,
                [
                    ('Today ', 'you', ' are You! '),
                    ('y you are ', 'You', '! That is '),
                    ('ve who is ', 'You', '-er than y'),
                    ('u-er than ', 'you', '!'),
                ],
            ),
            (
                re.compile("you(-er)?", flags=re.IGNORECASE),
                True,
                10,
                False,
                [
                    ('Today ', 'you', ' are You! '),
                    ('y you are ', 'You', '! That is '),
                    ('ve who is ', 'You-er', ' than you!'),
                    ('u-er than ', 'you', '!'),
                ],
            ),
            (
                re.compile("you(-er)?"),
                True,  # or False, it doesn't matter
                10,
                False,
                [
                    ('Today ', 'you', ' are You! '),
                    ('u-er than ', 'you', '!'),
                ],
            ),
        ],
    )
    def test_results(self, text, keyword, ignore_case, window_width, pad_context, exp):
        obs = list(
            extract.kwic.keyword_in_context(
                text,
                keyword,
                ignore_case=ignore_case,
                window_width=window_width,
                pad_context=pad_context,
            )
        )
        assert obs == exp



"""
Pipeline Components
-------------------

:mod:`textacy.text_stats.components`: Custom components to add to a spaCy language pipeline.
"""
# TODO: figure out why this breaks the code...
# from __future__ import annotations

import inspect
import logging
from typing import Collection, Optional, Union

from spacy.language import Language
from spacy.tokens import Doc

from . import api

LOGGER = logging.getLogger(__name__)


@Language.factory(
    "textacy_text_stats",
    default_config={"attrs": None},
    retokenizes=False,
)
def create_text_stats_component(
    nlp: Language, name: str, attrs: Optional[Union[str, Collection[str]]]
):
    return TextStatsComponent(attrs=attrs)


class TextStatsComponent:
    """
    A custom component to be added to a spaCy language pipeline that computes
    one, some, or all text stats for a parsed doc and sets the values
    as custom attributes on a :class:`spacy.tokens.Doc`.

    Add the component to a pipeline, *after* the parser and any subsequent components
    that modify the tokens/sentences of the doc (to be safe, just put it last)::

        >>> en = spacy.load("en_core_web_sm")
        >>> en.add_pipe("textacy_text_stats", last=True)

    Process a text with the pipeline and access the custom attributes via
    spaCy's underscore syntax::

        >>> doc = en(u"This is a test test someverylongword.")
        >>> doc._.n_words
        6
        >>> doc._.flesch_reading_ease
        73.84500000000001

    Specify which attributes of the :class:`textacy.text_stats.TextStats()`
    to add to processed documents::

        >>> en = spacy.load("en_core_web_sm")
        >>> en.add_pipe("textacy_text_stats", last=True, config={"attrs": "n_words"})
        >>> doc = en(u"This is a test test someverylongword.")
        >>> doc._.n_words
        6
        >>> doc._.flesch_reading_ease
        AttributeError: [E046] Can't retrieve unregistered extension attribute 'flesch_reading_ease'. Did you forget to call the `set_extension` method?

    Args:
        attr: If str, a single text stat to compute and set on a :obj:`Doc`;
            if Iterable[str], set multiple text stats; if None, *all* text stats
            are computed and set as extensions.

    See Also:
        :class:`textacy.text_stats.TextStats`
    """

    def __init__(self, attrs: Optional[Union[str, Collection[str]]] = None):
        self._set_attrs(attrs)
        for attr in self.attrs:
            # TODO: see if there's a better way to handle this
            # that doesn't involve clobbering existing property extensions
            Doc.set_extension(attr, default=None, force=True)
            LOGGER.debug('"%s" custom attribute added to `spacy.tokens.Doc`')

    def __call__(self, doc: Doc) -> Doc:
        ts = api.TextStats(doc)
        for attr in self.attrs:
            try:
                doc._.set(attr, getattr(ts, attr))
            except AttributeError:
                LOGGER.exception(
                    "`TextStats` class doesn't have '%s' attribute, so it can't "
                    "be set on this `Doc`. Check the attrs used to initialize "
                    "the `TextStatsComponent` in this pipeline for errors.",
                    attr,
                )
                raise
        return doc

    def _set_attrs(self, attrs: Optional[Union[str, Collection[str]]]):
        if attrs is None:
            self.attrs = tuple(
                name
                for name, _ in inspect.getmembers(
                    api.TextStats, lambda memb: not(inspect.isroutine(memb))
                )
                if not name.startswith("_")
            )
        elif isinstance(attrs, str):
            self.attrs = (attrs,)
        else:
            self.attrs = tuple(attrs)
